#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "board.h"
#include "rating.h"


/*------------------------------------------------------------------*/
/* O[oϐ                                                   */
/*------------------------------------------------------------------*/

/* R~ */
double komi[3];

/* Տ̍W */
int onboard_pos[PURE_BOARD_MAX];

/* XW */
int board_x[BOARD_MAX];

/* YW */
int board_y[BOARD_MAX];

/* 4ߖT̋_ */
int nb4_empty[PATTERN3_MAX];

/*  */
unsigned char eye[PATTERN3_MAX];

/*   */
unsigned char border_dis_x[BOARD_MAX];
unsigned char border_dis_y[BOARD_MAX];
unsigned char move_dis[PURE_BOARD_SIZE][PURE_BOARD_SIZE];


/*-------------------------------------------------------------------*/
/* ֐̐錾                                                        */
/*-------------------------------------------------------------------*/

/* ߖT̐̐ݒ */
static void InitializeNeighbor( void );

/* ̐ݒ */
static void InitializeEye( void );

/* E̔ */
static int CheckSuicide( struct stone *board, struct string *string, int color, int pos );

/* ͂܂ꂽ΂菜 */
static int RemoveStone( struct game *game, int str );

/* A̐VK쐬 */
static void CreateNewString( struct game *game, int color, int pos );

/* ̘͂Aɐ΂ǉ */
static void ExtendNeighborString( struct game *game, int str, int color, int pos );

/* A̍쐬 +  */
static void AssimilateNeighborString( struct game *game, int color, int pos );

/* A̓ */
static void AssimilateString( struct game *game, int str, int pos );

/* A_菜 */
static void RemoveLiberty( struct game *game, int str, int pos );

/* _̍XV */
static void UpdateLiberties( struct game *game, int str );

/* אڂA̒ǉ */
static void AddNeighbor( struct string *string, struct string_neighbors *string_neighbors, int str, int str2);

/* אڂA菜 */
static void RemoveNeighbor( struct string *string, struct string_neighbors *string_neighbors, int str, int n );


/*------------------------------------------------------------------*/
/*  (m + ݒ)                                   */
/*------------------------------------------------------------------*/
struct game *
InitializeGame( void )
{
	struct game *game = (struct game *)malloc(sizeof(struct game));

	memset(game, 0, sizeof(struct game));
	ClearGame(game);

	return game;
}


/*------------------------------------------------------------------*/
/* m                                                       */
/*------------------------------------------------------------------*/
struct game *
SetupGame( void )
{
	struct game *game = (struct game *)malloc(sizeof(struct game));

	memset(game, 0, sizeof(struct game));

	return game;
}


/*------------------------------------------------------------------*/
/* ݒ                                                         */
/*------------------------------------------------------------------*/
void
ClearGame( struct game *game )
{
	int x, y;

	memset(game->prisoner, 0, sizeof(int)*S_MAX);

	game->moves = 1;
	memset(game->record, 0, sizeof(struct move)*MAX_RECORDS);

	game->ko_pos = -1;
	game->ko_move = MAX_MOVES;

	memset(game->board, 0, sizeof(struct stone)*BOARD_MAX);
	for (y=0; y < BOARD_SIZE; y++) {
		for (x=0; x < OB_SIZE; x++) {
			game->board[POS(x,y)].color = S_OB;
			game->board[POS(BOARD_SIZE-1-x,y)].color = S_OB;
			game->board[POS(y,x)].color = S_OB;
			game->board[POS(y,BOARD_SIZE-1-x)].color = S_OB;
		}
	}

	memset(game->string, 0, sizeof(struct string)*MAX_STRINGS);
	memset(game->string_liberties, 0, sizeof(struct string_liberties)*MAX_STRINGS);
	memset(game->string_neighbors, 0, sizeof(struct string_neighbors)*MAX_STRINGS);
	game->next_string = 0;
	game->liberty_mark = 0;
	game->string_mark = 0;

	ClearPattern3( game->pat3 );
	ClearDiamond( game->dia );

	memset(game->candidates, 0, sizeof(int)*BOARD_MAX);
	for (y=BOARD_START; y<=BOARD_END; y++) {
		for (x=BOARD_START; x<=BOARD_END; x++) {
			game->candidates[POS(x,y)] = 1;
		}
	}

	ClearRate( game );
}


/*------------------------------------------------------------------*/
/* Rs[                                                           */
/*------------------------------------------------------------------*/
void
CopyGame( struct game *g2, struct game *g1 )
{

	memcpy(g2->prisoner, g1->prisoner, sizeof(int)*S_MAX);

	g2->moves = g1->moves;
	memcpy(g2->record, g1->record, sizeof(struct move)*MAX_RECORDS);

	g2->ko_pos = g1->ko_pos;
	g2->ko_move = g1->ko_move;

	memcpy(g2->board, g1->board, sizeof(struct stone)*BOARD_MAX);

	memcpy(g2->string, g1->string, sizeof(struct string)*MAX_STRINGS);
	memcpy(g2->string_liberties, g1->string_liberties,
		sizeof(struct string_liberties)*MAX_STRINGS);
	memcpy(g2->string_neighbors, g1->string_neighbors,
		sizeof(struct string_neighbors)*MAX_STRINGS);
	g2->next_string = g1->next_string;
	g2->liberty_mark = g1->liberty_mark;
	g2->string_mark = g1->string_mark;

	memcpy(g2->pat3, g1->pat3, sizeof(int)*BOARD_MAX);
	memcpy(g2->dia, g1->dia, sizeof(int)*BOARD_MAX);

	memcpy(g2->candidates, g1->candidates, sizeof(int)*BOARD_MAX);

	memcpy(g2->rate, g1->rate, sizeof(int)*BOARD_MAX);
	memcpy(g2->sum_rate_column, g1->sum_rate_column, sizeof(int)*BOARD_SIZE);
	g2->sum_rate = g1->sum_rate;
}


/*------------------------------------------------------------------*/
/*                                                        */
/*------------------------------------------------------------------*/
void
DoneGame( struct game *game )
{
	if (game) free(game);
}


/*------------------------------------------------------------------*/
/* 萔̐ݒ                                                       */
/*------------------------------------------------------------------*/
void
BoardConst( void )
{
	int x, y;
	int i;

	komi[0] = KOMI;
	komi[S_BLACK] = KOMI+1;
	komi[S_WHITE] = KOMI-1;

	i = 0;
	for (y=BOARD_START; y<=BOARD_END; y++) {
		for (x=BOARD_START; x<=BOARD_END; x++) {
			onboard_pos[i++] = POS(x, y);
			board_x[POS(x, y)] = x;
			board_y[POS(x, y)] = y;
		}
	}

	for (y=BOARD_START; y<=BOARD_END; y++) {
		for (x=BOARD_START; x <=(BOARD_START-1+((PURE_BOARD_SIZE+1)/2)); x++) {
			border_dis_x[POS(x,y)] = (unsigned char)(x-(OB_SIZE-1));
			border_dis_x[POS(BOARD_END+OB_SIZE-x,y)] = (unsigned char)(x-(OB_SIZE-1));
			border_dis_y[POS(y,x)] = (unsigned char)(x-(OB_SIZE-1));
			border_dis_y[POS(y,BOARD_END+OB_SIZE-x)] = (unsigned char)(x-(OB_SIZE-1));
		}
	}

	for (y=0; y < PURE_BOARD_SIZE; y++) {
		for (x=0; x < PURE_BOARD_SIZE; x++) {
			move_dis[x][y] = (unsigned char)(x + y + ((x>y)?x:y));
		}
	}

	InitializeNeighbor();
	InitializeEye();
}


/*------------------------------------------------------------------*/
/* ߖT̋_̐ݒ                                               */
/*------------------------------------------------------------------*/
static void
InitializeNeighbor( void )
{
	int i;
	int empty;

	for (i=0; i<PATTERN3_MAX; i++) {
		empty = 0;

		if (((i>>2) & 0x3) == S_EMPTY) empty++;
		if (((i>>6) & 0x3) == S_EMPTY) empty++;
		if (((i>>8) & 0x3) == S_EMPTY) empty++;
		if (((i>>12) & 0x3) == S_EMPTY) empty++;

		nb4_empty[i] = empty;
	}
}


/*------------------------------------------------------------------*/
/* ̐ݒ                                                         */
/*------------------------------------------------------------------*/
static void
InitializeEye( void )
{
	int i, j;
	int transp[8];
	static int eye_pat3[16] = { 
		0x5554, 0x5556, 0x5544, 0x5546, 0x1554,
		0x1556, 0x1544, 0x1546, 0x1564, 0x1146,
		0xFD54, 0xFD55, 0xFD56,
		0xFF74, 0xFF75, 0xFF76
	};

	eye[0x5555] = S_BLACK;
	eye[Pattern3Reverse(0x5555)] = S_WHITE;
	eye[0x1144] = S_BLACK;
	eye[Pattern3Reverse(0x1144)] = S_WHITE;

	for (i=0; i<16; i++) {
		Pattern3Transpose8( eye_pat3[i], transp );
		for (j=0; j<8; j++) {
			eye[transp[j]] = S_BLACK;
			eye[Pattern3Reverse(transp[j])] = S_WHITE;
		}
	}
}


/*------------------------------------------------------------------*/
/* @̔                                                     */
/*------------------------------------------------------------------*/
int
CheckLegal( struct game *game, int color, int pos )
{
	/* _ */
	if (game->board[pos].color != S_EMPTY) {
		return( FALSE );
	}

	/* E */
	if (nb4_empty[game->pat3[pos]] == 0
		&& CheckSuicide( game->board, game->string, color, pos) == TRUE) {
			return( FALSE );
	}

	/*  */
	if (game->ko_pos == pos && game->ko_move == (game->moves-1)) {
		return( FALSE );
	}

	return( TRUE );
}


/*------------------------------------------------------------------*/
/* @肩łȂ                                               */
/*------------------------------------------------------------------*/
int
CheckLegalAndEye( struct game *game, int color, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;

	/*  */
	if (eye[game->pat3[pos]] != color
		|| string[board[NORTH(pos)].string_number].liberties == 1
		|| string[board[ EAST(pos)].string_number].liberties == 1
		|| string[board[SOUTH(pos)].string_number].liberties == 1
		|| string[board[ WEST(pos)].string_number].liberties == 1){
			/* E */
			if (nb4_empty[game->pat3[pos]] == 0
				&& CheckSuicide( board, string, color, pos ) == TRUE) {
					return( FALSE );
			}

			/*  */
			if (game->ko_pos == pos
				&& game->ko_move == (game->moves-1)) {
					return( FALSE );
			}

			return( TRUE );
	}

	/* 肩珜O */
	game->candidates[pos] = 0;

	return( FALSE );
}


/*------------------------------------------------------------------*/
/* E̔                                                     */
/*------------------------------------------------------------------*/
static int
CheckSuicide( struct stone *board, struct string *string, int color, int pos )
{
	int other = OTHER_COLOR(color);

	/*  */
	if (board[NORTH(pos)].color == other) {
		if (string[board[NORTH(pos)].string_number].liberties == 1) {
			return( FALSE );
		}
	} else if (board[NORTH(pos)].color == color) {
		if (string[board[NORTH(pos)].string_number].liberties > 1) {
			return( FALSE );
		}
	}

	/* E */
	if (board[EAST(pos)].color == other) {
		if (string[board[EAST(pos)].string_number].liberties == 1) {
			return( FALSE );
		}
	} else if (board[EAST(pos)].color == color) {
		if (string[board[EAST(pos)].string_number].liberties > 1) {
			return( FALSE );
		}
	}

	/*  */
	if (board[SOUTH(pos)].color == other) {
		if (string[board[SOUTH(pos)].string_number].liberties == 1) {
			return( FALSE );
		}
	} else if ( board[SOUTH(pos)].color == color) {
		if (string[board[SOUTH(pos)].string_number].liberties > 1) {
			return( FALSE );
		}
	}

	/*  */
	if (board[WEST(pos)].color == other) {
		if (string[board[WEST(pos)].string_number].liberties == 1) {
			return( FALSE );
		}
	} else if (board[WEST(pos)].color == color) {
		if (string[board[WEST(pos)].string_number].liberties > 1) {
			return( FALSE );
		}
	}

	return( TRUE );
}


/*------------------------------------------------------------------*/
/* Ղɐ΂u                                                   */
/*------------------------------------------------------------------*/
void
SetStone( struct game *game, int color, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	int other = OTHER_COLOR(color);
	int str=0;
	int prisoner = 0;
	int neighbor_allies = 0;

	/*  */
	if (game->moves < MAX_RECORDS) {
		game->record[game->moves].color = color;
		game->record[game->moves].pos = pos;
	}

	/* pX */
	if (pos == PASS) {
		/* 萔 */
		game->moves++;
		return;
	}

	/* ΂u */
	board[pos].color = color;

	/* 肩珜O */
	game->candidates[pos] = 0;

	/* [g */
	game->sum_rate -= game->rate[pos];
	game->sum_rate_column[board_y[pos]] -= game->rate[pos];
	game->rate[pos] = 0;

	/* p^[̍XV */
	UpdatePattern3Stone( game->pat3, color, pos );
	UpdateDiamondStone( game->dia, color, pos );

	/* }[NXV */
	game->string_mark++;

	/*  */
	if (board[NORTH(pos)].color == color) {
		neighbor_allies++;
		str = board[NORTH(pos)].string_number;
		string[str].mark = game->string_mark;
	} else if (board[NORTH(pos)].color == other) {
		if (string[board[NORTH(pos)].string_number].liberties > 1) {
			RemoveLiberty( game, board[NORTH(pos)].string_number, pos );
			string[board[NORTH(pos)].string_number].mark = game->string_mark;
		} else {
			prisoner += RemoveStone( game, board[NORTH(pos)].string_number );
		}
	}

	/* E */
	if (board[EAST(pos)].color == color
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			neighbor_allies++;
			str = board[EAST(pos)].string_number;
			string[str].mark = game->string_mark;
	} else if (board[EAST(pos)].color == other
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			if (string[board[EAST(pos)].string_number].liberties > 1) {
				RemoveLiberty( game, board[EAST(pos)].string_number, pos );
				string[board[EAST(pos)].string_number].mark = game->string_mark;
			} else {
				prisoner += RemoveStone( game, board[EAST(pos)].string_number );
			}
	}

	/*  */
	if (board[SOUTH(pos)].color == color
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			neighbor_allies++;
			str = board[SOUTH(pos)].string_number;
			string[str].mark = game->string_mark;
	} else if (board[SOUTH(pos)].color == other
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			if (string[board[SOUTH(pos)].string_number].liberties > 1) {
				RemoveLiberty( game, board[SOUTH(pos)].string_number, pos );
				string[board[SOUTH(pos)].string_number].mark = game->string_mark;
			} else {
				prisoner += RemoveStone( game, board[SOUTH(pos)].string_number );
			}
	}

	/*  */
	if (board[WEST(pos)].color == color
		&& string[board[WEST(pos)].string_number].mark != game->string_mark) {
			neighbor_allies++;
			str = board[WEST(pos)].string_number;
			string[str].mark = game->string_mark;
	} else if (board[WEST(pos)].color == other
		&& string[board[WEST(pos)].string_number].mark != game->string_mark) {
			if (string[board[WEST(pos)].string_number].liberties > 1) {
				RemoveLiberty( game, board[WEST(pos)].string_number, pos );
				string[board[WEST(pos)].string_number].mark = game->string_mark;
			} else {
				prisoner += RemoveStone( game, board[WEST(pos)].string_number );
			}
	}

	/* AQn} */
	game->prisoner[color] += prisoner;

	/* A */
	if (neighbor_allies == 0) {
		CreateNewString( game, color, pos );
		/* ̔ */
		str = board[pos].string_number;
		if (prisoner == 1 && string[str].liberties == 1 && string[str].size == 1) {
			game->ko_move = game->moves;
			game->ko_pos = game->string_liberties[str].list[0];
		}
	} else if (neighbor_allies == 1) {
		ExtendNeighborString( game, str, color, pos );
	} else {
		AssimilateNeighborString( game, color, pos );
	}

	/* 萔 */
	game->moves++;
}


/*------------------------------------------------------------------*/
/* ͂܂ꂽ΂菜                                             */
/*------------------------------------------------------------------*/
static int
RemoveStone( struct game *game, int str )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	int k;
	int pos;

	pos = string[str].origin;
	do {
		/* p^[̍XV */
		UpdatePattern3Empty( game->pat3, pos );
		UpdateDiamondEmpty( game->dia, pos );

		/*  */
		board[pos].color = S_EMPTY;
		board[pos].string_number = 0;

		/* ɒǉ */
		game->candidates[pos] = 1;

		pos = board[pos].next_stone;
	} while (pos != string[str].origin);

	/* אڂAE_̍XV */
	for (k=0; k<string[str].neighbors; k++) {
		if (string[string_neighbors[str].list[k]].liberties < MAX_LIBERTIES){
			UpdateLiberties( game, string_neighbors[str].list[k] );
		}
		RemoveNeighbor( string, string_neighbors, string_neighbors[str].list[k], str );
	}

	/* A */
	string[str].mark = 0;

	/* ΂̐Ԃ */
	return string[str].size;
}


/*------------------------------------------------------------------*/
/* A̐VK쐬                                                     */
/*------------------------------------------------------------------*/
static void
CreateNewString( struct game *game, int color, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	struct string_liberties *string_liberties = game->string_liberties;
	int other = OTHER_COLOR(color);
	int str;

	/* AԍXV */
	do {
		game->next_string++;
		if (game->next_string > (MAX_STRINGS-1)) game->next_string = 1;
	} while (string[game->next_string].mark != 0);

	str = game->next_string;
	board[pos].string_number = str;
	board[pos].next_stone = pos;

	/* }[NXV */
	game->string_mark++;

	/*  */
	string[str].color = color;
	string[str].size = 1;
	string[str].origin = pos;
	string[str].liberties = 0;
	string[str].neighbors = 0;
	string[str].mark = game->string_mark;

	/*  */
	if (board[NORTH(pos)].color == other) {
		AddNeighbor(string, string_neighbors, str, board[NORTH(pos)].string_number);
		string[board[NORTH(pos)].string_number].mark = game->string_mark;
	} else if (board[NORTH(pos)].color == S_EMPTY) {
		string_liberties[str].list[string[str].liberties++] = NORTH(pos);
	}

	/* E */
	if (board[EAST(pos)].color == other
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[EAST(pos)].string_number);
			string[board[EAST(pos)].string_number].mark = game->string_mark;
	} else if (board[EAST(pos)].color == S_EMPTY) {
		string_liberties[str].list[string[str].liberties++] = EAST(pos);
	}

	/*  */
	if (board[SOUTH(pos)].color == other
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[SOUTH(pos)].string_number);
			string[board[SOUTH(pos)].string_number].mark = game->string_mark;
	} else if (board[SOUTH(pos)].color == S_EMPTY) {
		string_liberties[str].list[string[str].liberties++] = SOUTH(pos);
	}

	/*  */
	if (board[WEST(pos)].color == other
		&& string[board[WEST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[WEST(pos)].string_number);
	} else if (board[WEST(pos)].color == S_EMPTY) {
		string_liberties[str].list[string[str].liberties++] = WEST(pos);
	}
}


/*------------------------------------------------------------------*/
/* ̘͂Aɐ΂ǉ                                               */
/*------------------------------------------------------------------*/
static void
ExtendNeighborString( struct game *game, int str, int color, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	int other = OTHER_COLOR(color);
	int pos2;
	int k;

	string[str].size++;	

	/* Aԍ */
	board[pos].string_number = str;

	/* N */
	pos2 = string[str].origin;
	board[pos].next_stone = board[pos2].next_stone;
	board[pos2].next_stone = pos;

	if (pos < pos2) string[str].origin = pos;

	/* }[NXV */
	game->string_mark++;
	for (k=0; k<string[str].neighbors; k++) {
		string[string_neighbors[str].list[k]].mark = game->string_mark;
	}

	/*  */
	if (board[NORTH(pos)].color == other) {
		AddNeighbor(string, string_neighbors, str, board[NORTH(pos)].string_number);
		string[board[NORTH(pos)].string_number].mark = game->string_mark;
	}

	/* E */
	if (board[EAST(pos)].color == other
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[EAST(pos)].string_number);
			string[board[EAST(pos)].string_number].mark = game->string_mark;
	}

	/*  */
	if (board[SOUTH(pos)].color == other
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[SOUTH(pos)].string_number);
			string[board[SOUTH(pos)].string_number].mark = game->string_mark;
	}

	/*  */
	if (board[WEST(pos)].color == other
		&& string[board[WEST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[WEST(pos)].string_number);
	}

	/* _̍XV */
	UpdateLiberties( game, str );
}


/*------------------------------------------------------------------*/
/* A̍쐬 +                                                   */
/*------------------------------------------------------------------*/
static void
AssimilateNeighborString( struct game *game, int color, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	int other = OTHER_COLOR(color);
	int str;

	/* AԍXV */
	do {
		game->next_string++;
		if (game->next_string > (MAX_STRINGS-1)) game->next_string = 1;
	} while (string[game->next_string].mark != 0);

	str = game->next_string;
	board[pos].string_number = str;
	board[pos].next_stone = pos;

	/* }[NXV */
	game->string_mark++;

	/*  */
	string[str].color = color;
	string[str].size = 1;
	string[str].origin = pos;
	string[str].liberties = 0;
	string[str].neighbors = 0;
	string[str].mark = game->string_mark;

	/*  */
	if (board[NORTH(pos)].color == color) {
		AssimilateString( game, str, NORTH(pos) );
	} else if (board[NORTH(pos)].color == other) {
		AddNeighbor(string, string_neighbors, str, board[NORTH(pos)].string_number);
		string[board[NORTH(pos)].string_number].mark = game->string_mark;
	}

	/* E */
	if (board[EAST(pos)].color == color
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			AssimilateString( game, str, EAST(pos) );
	} else if (board[EAST(pos)].color == other
		&& string[board[EAST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[EAST(pos)].string_number);
			string[board[EAST(pos)].string_number].mark = game->string_mark;
	}

	/*  */
	if (board[SOUTH(pos)].color == color
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			AssimilateString( game, str, SOUTH(pos) );
	} else if (board[SOUTH(pos)].color == other
		&& string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[SOUTH(pos)].string_number);
			string[board[SOUTH(pos)].string_number].mark = game->string_mark;
	}

	/*  */
	if (board[WEST(pos)].color == color
		&& string[ board[WEST(pos)].string_number].mark != game->string_mark) {
			AssimilateString( game, str, WEST(pos));
	} else if (board[WEST(pos)].color == other
		&& string[ board[WEST(pos)].string_number].mark != game->string_mark) {
			AddNeighbor(string, string_neighbors, str, board[WEST(pos)].string_number);
	}

	/* _̍XV */
	UpdateLiberties( game, str );
}


/*------------------------------------------------------------------*/
/* A̓                                                         */
/*------------------------------------------------------------------*/
static void
AssimilateString( struct game *game, int str, int pos )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	int k, t;
	int last, pos2;
	int str2 = board[pos].string_number;

	string[str].size += string[str2].size;

	/* Aԍ */
	pos = string[str2].origin;
	do {
		board[pos].string_number = str;
		last = pos;
		pos = board[pos].next_stone;
	} while (pos != string[str2].origin);

	/* N */
	pos2 = string[str].origin;
	board[last].next_stone = board[pos2].next_stone;
	board[pos2].next_stone = string[str2].origin;

	if (string[str2].origin < pos2) {
		string[str].origin = string[str2].origin;
	}

	/* אڂA */
	for (k = 0; k < string[str2].neighbors; k++) {
		t = string_neighbors[str2].list[k];
		RemoveNeighbor( string, string_neighbors, t, str2 );
		if (string[t].mark != game->string_mark) {
			string_neighbors[t].list[string[t].neighbors++] = str;
			string_neighbors[str].list[string[str].neighbors++] = t;
			string[t].mark = game->string_mark;
		}
	}

	/* jAԍ */
	string[str2].mark = 0;
}


/*------------------------------------------------------------------*/
/* אڂAǉ                                                 */
/*------------------------------------------------------------------*/
static void
AddNeighbor( struct string *string, struct string_neighbors *string_neighbors, int str, int str2 )
{
	string_neighbors[str].list[string[str].neighbors++] = str2;
	string_neighbors[str2].list[string[str2].neighbors++] = str;
}


/*------------------------------------------------------------------*/
/* אڂA菜  (Astr  Astr2 菜)             */
/*------------------------------------------------------------------*/
static void
RemoveNeighbor( struct string *string, struct string_neighbors *string_neighbors, int str, int str2 )
{
	int k;

	for (k = 0; k < string[str].neighbors; k++) {
		if (string_neighbors[str].list[k] == str2) {
			string_neighbors[str].list[k] = string_neighbors[str].list[--string[str].neighbors];
			break;
		}
	}
}


/*------------------------------------------------------------------*/
/* A_菜  (Astr  _pos 菜)            */
/*------------------------------------------------------------------*/
static void
RemoveLiberty( struct game *game, int str, int pos )
{
	struct string *string = &game->string[str];
	struct string_liberties *string_liberties = &game->string_liberties[str];
	int k;

	if (string->liberties >= MAX_LIBERTIES) {
		UpdateLiberties( game, str );
	} else {
		for (k = 0; k < string->liberties; k++) {
			if (string_liberties->list[k] == pos) {
				string_liberties->list[k] = string_liberties->list[--string->liberties];
				break;
			}
		}
	}

	if (string->liberties == 1) {
		game->candidates[string_liberties->list[0]] = 1;
	}
}


/*------------------------------------------------------------------*/
/* _̍XV                                                       */
/*------------------------------------------------------------------*/
static void
UpdateLiberties( struct game *game, int str )
{
	struct stone *board = game->board;
	struct string *string = &game->string[str];
	struct string_liberties *string_liberties = &game->string_liberties[str];
	int pos;

	/*  */
	string->liberties = 0;

	/* }[NXV */
	game->liberty_mark++;

	pos = string->origin;
	do {
		if(nb4_empty[game->pat3[pos]] != 0){
			if (board[NORTH(pos)].color == S_EMPTY
				&& board[NORTH(pos)].mark != game->liberty_mark) {
					string_liberties->list[string->liberties++] = NORTH(pos);
					if (string->liberties >= MAX_LIBERTIES) break;
					board[NORTH(pos)].mark = game->liberty_mark;
			}

			if (board[EAST(pos)].color == S_EMPTY
				&& board[EAST(pos)].mark != game->liberty_mark) {
					string_liberties->list[string->liberties++] = EAST(pos);
					if (string->liberties >= MAX_LIBERTIES) break;
					board[EAST(pos)].mark = game->liberty_mark;
			}

			if (board[SOUTH(pos)].color == S_EMPTY
				&& board[SOUTH(pos)].mark != game->liberty_mark) {
					string_liberties->list[string->liberties++] = SOUTH(pos);
					if (string->liberties >= MAX_LIBERTIES) break;
					board[SOUTH(pos)].mark = game->liberty_mark;
			}

			if (board[WEST(pos)].color == S_EMPTY
				&& board[WEST(pos)].mark != game->liberty_mark) {
					string_liberties->list[string->liberties++] = WEST(pos);
					if (string->liberties >= MAX_LIBERTIES) break;
					board[WEST(pos)].mark = game->liberty_mark;
			}
		}
		pos = board[pos].next_stone;
	} while (pos != string->origin);

	if (string->liberties == 1) {
		game->candidates[string_liberties->list[0]] = 1;
	}
}


/*------------------------------------------------------------------*/
/* ňvZ                                                         */
/*------------------------------------------------------------------*/
int
BoardScore( struct game *game )
{
	struct stone *board = game->board;
	int i;
	int pos;
	int color;
	int scores[S_MAX];

	memset(scores, 0, sizeof(scores));

	for (i=0; i<PURE_BOARD_MAX; i++) {
		pos = onboard_pos[i];
		color = board[pos].color;
		if (color == S_EMPTY) color = eye[game->pat3[pos]];
		scores[color]++;
	}

	/* |Ԃ(R~Ȃ) */
	return( scores[S_BLACK] - scores[S_WHITE] );
}


/*------------------------------------------------------------------*/
/* Ղ\                                                       */
/*------------------------------------------------------------------*/
void
DisplayBoard( struct game *game )
{
	static char stone[] = { '+', '@', 'O', '?' };
	int x, y;

	fprintf(stderr, "### Board ### move=%d\n", game->moves);
	fprintf(stderr, "    [");
	for (y=1; y<=PURE_BOARD_SIZE; y++) {
		fprintf(stderr, "%2d", y);
	}
	fprintf(stderr, "]\n");
	for (y=BOARD_START; y<=BOARD_END; y++) {
		fprintf(stderr, "[%2d] ", y-(OB_SIZE-1));
		for (x=BOARD_START; x<=BOARD_END; x++) {
			fprintf(stderr, " %c", stone[game->board[POS(x, y)].color]);
		}
		fprintf(stderr, "\n");
	}
	fprintf(stderr, "\n");
}